﻿#if UNITY_EDITOR
using UnityEngine;
using UnityEditor;
using System;
using System.Collections;
using System.Collections.Generic;
using UPLib;

//TODO: light + AmbientSound actors, triggerlights (w/o triggering)
//TODO: should Unreal tags be turned into Unity tags?

[CustomEditor(typeof(UnLevel))]
public class UnLevelInspector : Editor
{
	SerializedProperty levelPathProp;
	SerializedProperty levelScaleProp;
	SerializedProperty generateLightmapUVsProp;
	SerializedProperty lightmapSeamFixProp;
	SerializedProperty noFakeBackdropPolysProp;

	SerializedProperty texturesPathProp;
	SerializedProperty soundsPathProp;

	SerializedProperty materialNormalProp;
	SerializedProperty materialMaskedProp;
	SerializedProperty materialModulatedProp;
	SerializedProperty materialTranslucentProp;
	SerializedProperty materialUnlitProp;

	void OnEnable () {
		// Setup the SerializedProperties.
		levelPathProp  = serializedObject.FindProperty("levelPath");
		levelScaleProp = serializedObject.FindProperty("levelScale");
		generateLightmapUVsProp = serializedObject.FindProperty("generateLightmapUVs");
		lightmapSeamFixProp = serializedObject.FindProperty("lightmapSeamFix");
		noFakeBackdropPolysProp = serializedObject.FindProperty("noFakeBackdropPolys");

		texturesPathProp = serializedObject.FindProperty("texturesPath");
		soundsPathProp = serializedObject.FindProperty("soundsPath");

		materialNormalProp = serializedObject.FindProperty("materialNormal");
		materialMaskedProp = serializedObject.FindProperty("materialMasked");
		materialModulatedProp = serializedObject.FindProperty("materialModulated");
		materialTranslucentProp = serializedObject.FindProperty("materialTranslucent");
		materialUnlitProp = serializedObject.FindProperty("materialUnlit");

		// Make sure the materials are set up.
		if(materialNormalProp.objectReferenceValue      == null)
			materialNormalProp.objectReferenceValue      = AssetDatabase.LoadAssetAtPath<Material>("Assets/Scripts/UE1ToUnity/Materials/Normal.mat"     );
		if(materialMaskedProp.objectReferenceValue      == null)
			materialMaskedProp.objectReferenceValue      = AssetDatabase.LoadAssetAtPath<Material>("Assets/Scripts/UE1ToUnity/Materials/Masked.mat"     );
		if(materialModulatedProp.objectReferenceValue   == null)
			materialModulatedProp.objectReferenceValue   = AssetDatabase.LoadAssetAtPath<Material>("Assets/Scripts/UE1ToUnity/Materials/Modulated.mat"  );
		if(materialTranslucentProp.objectReferenceValue == null)
			materialTranslucentProp.objectReferenceValue = AssetDatabase.LoadAssetAtPath<Material>("Assets/Scripts/UE1ToUnity/Materials/Translucent.mat");
		if(materialUnlitProp.objectReferenceValue       == null)
			materialUnlitProp.objectReferenceValue       = AssetDatabase.LoadAssetAtPath<Material>("Assets/Scripts/UE1ToUnity/Materials/Unlit.mat"      );
		serializedObject.ApplyModifiedProperties ();
	}

	public override void OnInspectorGUI()
	{
		EditorGUILayout.LabelField (levelPathProp.stringValue);
		EditorGUILayout.BeginHorizontal ();
		if (GUILayout.Button ("Import Unreal Engine 1 map")) {
			var initialDir = levelPathProp.stringValue != "" ? System.IO.Path.GetDirectoryName(levelPathProp.stringValue) : "";
			// TODO: Prefer storing relative path
			var path = EditorUtility.OpenFilePanel ("Import Unreal Engine 1 map", initialDir, "unr");
			if (path.Length != 0) {
				// import the map...
				levelPathProp.stringValue = path;
				ImportMap(path);
			}
		}
		if (GUILayout.Button ("Reload map")) {
			if(levelPathProp.stringValue != "")
				ImportMap(levelPathProp.stringValue);
		}
		EditorGUILayout.EndHorizontal ();
		levelScaleProp.floatValue         = EditorGUILayout.FloatField ("Import scale", levelScaleProp.floatValue);
		generateLightmapUVsProp.boolValue = EditorGUILayout.BeginToggleGroup ("Generate lightmap UVs", generateLightmapUVsProp.boolValue);
		lightmapSeamFixProp.floatValue    = EditorGUILayout.Slider ("Lightmap seams fix", lightmapSeamFixProp.floatValue,4,64);
		EditorGUILayout.EndToggleGroup ();

		noFakeBackdropPolysProp.boolValue = EditorGUILayout.Toggle ("No FakeBackdrop polys", noFakeBackdropPolysProp.boolValue);

		EditorGUILayout.LabelField ("Root directories");
		texturesPathProp.stringValue = EditorGUILayout.TextField("   Textures", texturesPathProp.stringValue);
		soundsPathProp.stringValue   = EditorGUILayout.TextField("   Sounds"  , soundsPathProp.stringValue  );

		EditorGUILayout.LabelField ("Materials");
		materialNormalProp.objectReferenceValue     =EditorGUILayout.ObjectField("   Normal"     ,materialNormalProp.objectReferenceValue     ,typeof(Material),false);
		materialMaskedProp.objectReferenceValue     =EditorGUILayout.ObjectField("   Masked"     ,materialMaskedProp.objectReferenceValue     ,typeof(Material),false);
		materialModulatedProp.objectReferenceValue  =EditorGUILayout.ObjectField("   Modulated"  ,materialModulatedProp.objectReferenceValue  ,typeof(Material),false);
		materialTranslucentProp.objectReferenceValue=EditorGUILayout.ObjectField("   Translucent",materialTranslucentProp.objectReferenceValue,typeof(Material),false);
		materialUnlitProp.objectReferenceValue      =EditorGUILayout.ObjectField("   Unlit"      ,materialUnlitProp.objectReferenceValue      ,typeof(Material),false);

		// Apply changes to the serializedProperty - always do this at the end of OnInspectorGUI.
		serializedObject.ApplyModifiedProperties ();
	}

	void ImportMap(string file)
	{
		var importer = CreateInstance<LevelImporting> ();
		importer.srcFile = file;
		importer.scale = levelScaleProp.floatValue;
		importer.generateLightmapUVs = generateLightmapUVsProp.boolValue;
		importer.lightmapSeamFix = lightmapSeamFixProp.floatValue;
		importer.noFakeBackdropPolys = noFakeBackdropPolysProp.boolValue;
		importer.texturesPath = texturesPathProp.stringValue;
		importer.soundsPath   = soundsPathProp.stringValue  ;
		importer.materialModulated   = materialModulatedProp.objectReferenceValue   as Material;
		importer.materialTranslucent = materialTranslucentProp.objectReferenceValue as Material;
		importer.materialMasked      = materialMaskedProp.objectReferenceValue      as Material;
		importer.materialUnlit       = materialUnlitProp.objectReferenceValue       as Material;
		importer.materialNormal      = materialNormalProp.objectReferenceValue      as Material;
		importer.Import ((UnLevel)target);
	}
}

public class LevelImporting : ScriptableObject
{
	public string  srcFile;
	public float scale = 1f/64f;
	public bool  generateLightmapUVs = true;
	public float lightmapSeamFix     = 35f ;
	public bool  noFakeBackdropPolys = true;
	public string texturesPath = "Level textures";
	public string soundsPath = "Level sounds";
	public Material materialModulated  ;
	public Material materialTranslucent;
	public Material materialMasked     ;
	public Material materialUnlit      ;
	public Material materialNormal     ;
	
	UnLevel root;
	bool finishedImport;

	PackageClasses.Package pkg;
	PackageClasses.ExportTableItem[] zoneInfos = new PackageClasses.ExportTableItem[64];
	List<Zone>  zones  = new List<Zone >();
	List<Mover> movers = new List<Mover>();
	
	HashSet<string> thrownMissingTextureWarnings = new HashSet<string>(); //Don't whine about the same texture twice.
	//TODO maybe: List<GameObject> objectPool; //Blanking objects instead of removing them preserves references to them.
	UnActor[] oldObjects;

	class PolyGroup
	{
		public int textureNum = 0;
		public string textureName = "";
		public Texture2D texture  = null;
		public Material  material = null;
		public int polyFlags = 0;
		public bool belongsToMover = false;
		
		public List<Vector3>  verts   = new List<Vector3 >();
		public List<Vector2>  uv      = new List<Vector2 >();
		public List<Vector3>  normals = new List<Vector3 >();
		public List<Triangle> tris    = new List<Triangle>();
	}
	
	class Zone
	{
		public int iZone = 0;
		public PackageClasses.ExportTableItem zoneInfo;
		public GameObject gameObject = null;
		
		public List<PolyGroup> polyGroups = new List<PolyGroup>();
	}
	
	class Mover
	{
		public string name = "";
		public GameObject gameObject = null;
		public List<PolyGroup> polyGroups = new List<PolyGroup>();
	}

	const int PF_Invisible			= 0x0000001;	// Poly is invisible.
	const int PF_Masked				= 0x0000002;	// Poly should be drawn masked.
	const int PF_Translucent	 	= 0x0000004;	// Poly is transparent.
	const int PF_NotSolid			= 0x0000008;	// Poly is not solid, doesn't block.
	const int PF_Environment   		= 0x0000010;	// Poly should be drawn environment mapped.
	const int PF_ForceViewZone		= 0x0000010;	// Force current iViewZone in OccludeBSP (reuse Environment flag)
	const int PF_Semisolid	  		= 0x0000020;	// Poly is semi-solid = collision solid, Csg nonsolid.
	const int PF_Modulated 			= 0x0000040;	// Modulation transparency.
	const int PF_FakeBackdrop		= 0x0000080;	// Poly looks exactly like backdrop.
	const int PF_TwoSided			= 0x0000100;	// Poly is visible from both sides.
	const int PF_AutoUPan		 	= 0x0000200;	// Automatically pans in U direction.
	const int PF_AutoVPan 			= 0x0000400;	// Automatically pans in V direction.
	const int PF_NoSmooth			= 0x0000800;	// Don't smooth textures.
	const int PF_SpecialPoly		= 0x0001000;	// Game-specific poly-level render control (reuse BigWavy flag)
	const int PF_SmallWavy			= 0x0002000;	// Small wavy pattern (for water/enviro reflection).
	const int PF_Flat				= 0x0004000;	// Flat surface.
	const int PF_LowShadowDetail	= 0x0008000;	// Low detaul shadows.
	const int PF_NoMerge			= 0x0010000;	// Don't merge poly's nodes before lighting when rendering.
	const int PF_AlphaBlend      	= 0x0020000;	// 227-only
	const int PF_DirtyShadows		= 0x0040000;	// Dirty shadows.
	const int PF_BrightCorners		= 0x0080000;	// Brighten convex corners.
	const int PF_SpecialLit			= 0x0100000;	// Only speciallit lights apply to this poly.
	const int PF_NoBoundRejection	= 0x0200000;	// Disable bound rejection in OccludeBSP (reuse Gourard flag)
	const int PF_Unlit				= 0x0400000;	// Unlit.
	const int PF_HighShadowDetail	= 0x0800000;	// High detail shadows.
	const int PF_Portal				= 0x4000000;	// Portal between iZones.
	const int PF_Mirrored			= 0x8000000;	// Reflective surface.
	
	const int relevantFlags = 139601887; // PolyFlags that we care about.
	
	Texture2D LoadTexture(string name)
	{
		var firstDotPos = name.IndexOf (".");
		if(firstDotPos!=-1)
		{
			name = name.Substring (0, firstDotPos) + "/Textures/" + name.Substring (firstDotPos+1);
			name = name.Replace (".", "/");
		}
		
		name = "Assets/" + texturesPath + "/" + name;
		
		if (System.IO.File.Exists (name + ".bmp"))
			name += ".bmp";
		else if(System.IO.File.Exists (name + ".png"))
			name += ".png";
		else if(System.IO.File.Exists (name + ".jpg"))
			name += ".jpg";
		else if(System.IO.File.Exists (name + ".dds"))
			name += ".dds";
		
		return AssetDatabase.LoadAssetAtPath<Texture2D>(name);
	}

	AudioClip LoadSound(string name)
	{
		name = name.Replace("Sound'", "").Replace ("'","");

		var firstDotPos = name.IndexOf (".");
		if(firstDotPos!=-1)
		{
			name = name.Substring (0, firstDotPos) + "/Sounds/" + name.Substring (firstDotPos+1);
			name = name.Replace (".", "/");
		}
		
		name = "Assets/" + soundsPath + "/" + name;
		
		if (System.IO.File.Exists (name + ".wav"))
			name += ".wav";
		else if(System.IO.File.Exists (name + ".ogg"))
			name += ".ogg";
		else if(System.IO.File.Exists (name + ".mp3"))
			name += ".mp3";

		return AssetDatabase.LoadAssetAtPath<AudioClip>(name);
	}
	
	private void WarnMissingTexture(string name)
	{
		if (!thrownMissingTextureWarnings.Contains (name)) {
			Debug.LogWarning ("Missing texture " + name + ". UV maps will be broken.");
			thrownMissingTextureWarnings.Add (name);
		}
	}
	
	PolyGroup GetPolyGroup(int texture, int iZone, int polyFlags)
	{
		foreach (var zone in zones) {
			if(zone.iZone == iZone)
			{
				foreach (var group in zone.polyGroups)
					if(group.textureNum == texture && group.polyFlags == polyFlags)
						return group;
				
				// group not found?
				var newGroup = new PolyGroup();
				newGroup.textureNum = texture;
				newGroup.textureName = pkg.GetFullObjectName (texture, false);
				newGroup.texture = LoadTexture(newGroup.textureName);
				if (newGroup.texture == null)
					WarnMissingTexture(newGroup.textureName);
				newGroup.polyFlags = polyFlags;
				zone.polyGroups.Add (newGroup);
				return newGroup;
			}
		}
		
		//zone and group not found?
		var newZone = new Zone ();
		newZone.iZone = iZone;
		newZone.zoneInfo = zoneInfos[iZone];
		zones.Add (newZone);
		
		var newGroup2 = new PolyGroup();
		newGroup2.textureNum = texture;
		newGroup2.textureName = pkg.GetFullObjectName (texture, false);
		newGroup2.texture = LoadTexture(newGroup2.textureName);
		if (newGroup2.texture == null)
			WarnMissingTexture(newGroup2.textureName);
		newGroup2.polyFlags = polyFlags;
		newZone.polyGroups.Add (newGroup2);
		return newGroup2;
	}
	
	PolyGroup GetMoverPolyGroup(int texture, int polyFlags)
	{
		var mover = movers [movers.Count - 1]; //TODO: this is ugly. refactor it...somehow
		
		foreach (var group in mover.polyGroups)
			if(group.textureNum == texture && group.polyFlags == polyFlags)
				return group;
		
		// group not found?
		var newGroup = new PolyGroup();
		newGroup.textureNum = texture;
		newGroup.textureName = pkg.GetFullObjectName (texture, false);
		newGroup.texture = LoadTexture(newGroup.textureName);
		if (newGroup.texture == null)
			WarnMissingTexture(newGroup.textureName);
		newGroup.polyFlags = polyFlags;
		mover.polyGroups.Add (newGroup);
		return newGroup;
	}
	
	struct Polygon
	{
		public bool belongsToMover;
		
		public int Texture;
		public int iZone;
		
		public int Flags;
		public Vector3 Origin;
		public Vector3 Normal;
		public Vector3 TextureU;
		public Vector3 TextureV;
		public short UPan, VPan;
		public Vector3[] Verts;
	}
	
	struct Triangle
	{
		public int iVert0, iVert1, iVert2;
		public int Texture, iZone;
	}
	
	static Color HSVToRGB(float LBrightness, float LHue, float LSaturation)
	{
		float r=0, g=0, b=0;
		const float unit = 256f / 6f;
		
		float hueFrac = (LHue / unit) % 1f;
		
		if(LHue < unit) {
			r = 1;
			g = hueFrac;
			b = 0;
		} else if(LHue < unit*2) {
			r = 1 - hueFrac;
			g = 1;
			b = 0;
		} else if(LHue < unit*3) {
			r = 0;
			g = 1;
			b = hueFrac;
		} else if(LHue < unit*4) {
			r = 0;
			g = 1 - hueFrac;
			b = 1;
		} else if(LHue < unit*5) {
			r = hueFrac;
			g = 0;
			b = 1;
		} else {
			r = 1;
			g = 0;
			b = 1 - hueFrac;
		}
		
		LSaturation /= 255f;
		
		r = Mathf.Lerp (r,1,LSaturation);
		g = Mathf.Lerp (g,1,LSaturation);
		b = Mathf.Lerp (b,1,LSaturation);
		
		r *= LBrightness / 255f;
		g *= LBrightness / 255f;
		b *= LBrightness / 255f;
		
		r = Mathf.Clamp01 (r);
		g = Mathf.Clamp01 (g);
		b = Mathf.Clamp01 (b);

		return new Color (r, g, b);
	}
	
	string GetLevelTitle(PackageClasses.Package level)
	{
		foreach (var exp in level.ExportTable) {
			if (Utils.StrEq(exp.ClassNameStr, "LevelInfo"))
			{
				var props = exp.GetProperties ();
				
				var title = props["Title"];
				if(title != "")
				{
					title = title.Substring(1,title.Length-2); //remove quotes
					return title;
				}
				else
					return "Unreal Level";
			}
		}
		return "Unreal Level";
	}

	private static Vector3 RotatorToUnity(Rotator r)
	{
		return new Vector3 (-r.Pitch*360f/65536f, r.Yaw*360f/65536f + 90f, -r.Roll*360f/65536f);
	}
	
	// Convert Unreal rotator to Unity eulerAngles.
	private static Vector3 ConvertRotator(Vector3 rot)
	{
		rot += new Vector3(0,-90,0);
		float temp;
		temp = rot.x;
		rot.x = rot.z;
		rot.z = -temp;
		return rot;
	}
	
	private static Vector3 VectorToUnity(Vector v)
	{
		return new Vector3 (v.X, v.Y, v.Z);
	}
	
	private static Vector3 SwapAxes(Vector3 v)
	{
		return new Vector3 (v.x, v.z, -v.y);
	}
	
	void AddBrushPolygon(Polygon poly)
	{
		PolyGroup group;
		
		if (poly.belongsToMover)
			group = GetMoverPolyGroup (poly.Texture, poly.Flags & relevantFlags);
		else
			group = GetPolyGroup (poly.Texture, poly.iZone, poly.Flags & relevantFlags);
		
		float UDot = Vector3.Dot (poly.TextureU, poly.Origin);
		float VDot = Vector3.Dot (poly.TextureV, poly.Origin);
		
		int texW, texH;
		if (group.texture != null) {
			texW = group.texture.width;
			texH = group.texture.height;
		} else {
			texW = 128;
			texH = 128;
		}
		
		int iFirstVert = group.verts.Count;
		
		for (int i=0; i<poly.Verts.Length; i++) {
			group.verts.Add (poly.Verts[i]);
			group.normals.Add (poly.Normal);
			
			float U = Vector3.Dot (poly.TextureU, poly.Verts[i]);
			float V = Vector3.Dot (poly.TextureV, poly.Verts[i]);
			
			Vector2 curUV;
			curUV.x = (U-UDot+poly.UPan)/texW;
			curUV.y = (V-VDot+poly.VPan)/texH;
			curUV.y = 1-curUV.y;
			group.uv.Add (curUV);
		}
		
		for (int i=0; i<poly.Verts.Length-2; i++) {
			Triangle t;
			
			t.iVert0 = iFirstVert;
			t.iVert1 = iFirstVert+i+1;
			t.iVert2 = iFirstVert+i+2;
			t.Texture = poly.Texture;
			t.iZone = poly.iZone;
			
			group.tris.Add (t);
		}
	}
	
	void FinishZone(Zone zone)
	{
		if (zone.zoneInfo == null)
			Debug.LogWarning ("Bad zoning. Rebuild the map and import it again.");

		string zoneInfoName = zone.zoneInfo == null ? "LevelInfo" : zone.zoneInfo.ObjectNameStr;

		var zoneObj = new GameObject (zoneInfoName + (zoneInfoName.StartsWith ("LevelInfo",StringComparison.OrdinalIgnoreCase) ? " " + zone.iZone : ""));
		zoneObj.transform.SetParent(root.transform, false);
		zoneObj.isStatic = true;
		zone.gameObject = zoneObj;
		
		var zoneCmp = zoneObj.AddComponent<UnZone>();
		zoneCmp.detached = false;
		zoneCmp.originalLevel = root.gameObject;
		zoneCmp.originalName = zone.zoneInfo.ObjectNameStr;

		foreach(var prop in zone.zoneInfo.GetProperties().ToStringList())
			zoneCmp.properties.Add(prop.Name, prop.Value);

		zone.polyGroups.Sort ((x, y) => x.textureName.CompareTo (y.textureName));
		
		foreach (var group in zone.polyGroups)
			FinishGroup (zone.gameObject, group);
	}
	
	void FinishGroup(GameObject parent, PolyGroup group)
	{
		var groupObj = new GameObject("Group " + group.textureName + ((group.polyFlags != 0) ? " " + group.polyFlags.ToString("X") : ""));
		groupObj.transform.SetParent(parent.transform, false);
		groupObj.isStatic = parent.isStatic;

		if (!group.belongsToMover) {
			var actorCmp = groupObj.AddComponent<UnActor> ();
			actorCmp.detached = false;
			actorCmp.originalLevel = root.gameObject;
			actorCmp.originalName = groupObj.name;
		}
		
		var newVertices = new Vector3[group.verts.Count];
		for (int i=0; i<group.verts.Count; i++) {
			newVertices[i] = SwapAxes(group.verts[i])*scale;
		}
		
		var newUVs = group.uv.ToArray ();
		
		var newNormals = new Vector3[group.normals.Count];
		for (int i=0; i<group.normals.Count; i++) {
			newNormals[i] = SwapAxes(group.normals[i]);
		}
		
		bool makeTwoSided = ((group.polyFlags & PF_TwoSided) != 0) && ((group.polyFlags & (PF_Translucent | PF_Modulated)) == 0);
		int[] newTriangles;
		
		if(makeTwoSided)
			newTriangles = new int[group.tris.Count*6];
		else
			newTriangles = new int[group.tris.Count*3];
		
		for (int i=0; i<group.tris.Count; i++) {
			if(makeTwoSided)
			{
				newTriangles[i*6  ] = group.tris[i].iVert0;
				newTriangles[i*6+1] = group.tris[i].iVert1;
				newTriangles[i*6+2] = group.tris[i].iVert2;
				newTriangles[i*6+3] = group.tris[i].iVert2;
				newTriangles[i*6+4] = group.tris[i].iVert1;
				newTriangles[i*6+5] = group.tris[i].iVert0;
			}
			else
			{
				newTriangles[i*3  ] = group.tris[i].iVert0;
				newTriangles[i*3+1] = group.tris[i].iVert1;
				newTriangles[i*3+2] = group.tris[i].iVert2;
			}
		}
		
		var mesh = new Mesh();
		mesh.name = "Mesh " + group.textureName + ((group.polyFlags != 0) ? " " + group.polyFlags.ToString("X") : "");
		mesh.vertices = newVertices;
		mesh.uv = newUVs;
		mesh.normals = newNormals;
		mesh.triangles = newTriangles;
		
		// Do we need this?
		mesh.RecalculateNormals ();
		
		// Without this, lightmaps will be a glitched mess
		if (generateLightmapUVs) {
			var unwrapParameters = new UnwrapParam ();
			unwrapParameters.angleError = 0.08f;
			unwrapParameters.areaError = 0.15f;
			unwrapParameters.hardAngle = 88f;
			unwrapParameters.packMargin = lightmapSeamFix / 1024f;
			Unwrapping.GenerateSecondaryUVSet (mesh, unwrapParameters);
		}
		
		var mf = groupObj.AddComponent<MeshFilter> ();
		mf.mesh = mesh;
		
		if ((group.polyFlags & (PF_Invisible | PF_FakeBackdrop)) == 0) {
			var mr = groupObj.AddComponent<MeshRenderer> ();
			
			Material mat;
			if     ((group.polyFlags & PF_Modulated  ) != 0)
				mat = materialModulated;
			else if((group.polyFlags & PF_Translucent) != 0)
				mat = materialTranslucent;
			else if((group.polyFlags & PF_Masked     ) != 0)
				mat = materialMasked;
			else if((group.polyFlags & PF_Unlit      ) != 0)
				mat = materialUnlit;
			else
				mat = materialNormal;
			
			mr.sharedMaterial = Instantiate(mat);
			if(group.texture != null)
				mr.sharedMaterial.mainTexture = group.texture;
			
			if ((group.polyFlags & PF_NotSolid) != 0)
				mr.shadowCastingMode = UnityEngine.Rendering.ShadowCastingMode.Off;
		}
		
		if ((group.polyFlags & (PF_AutoUPan | PF_AutoVPan | PF_SmallWavy)) != 0) {
			var pannerCmp = groupObj.AddComponent<UnTexPanner>();

			float zoneUPanSpeed = 1;
			float zoneVPanSpeed = 1;
			var zone = groupObj.transform.parent.GetComponent<UnZone>();
			if(zone != null)
			{
				zoneUPanSpeed = float.Parse(zone.GetProp("TexUPanSpeed","1"), System.Globalization.CultureInfo.InvariantCulture);
				zoneVPanSpeed = float.Parse(zone.GetProp("TexVPanSpeed","1"), System.Globalization.CultureInfo.InvariantCulture);
			}

			if((group.polyFlags & PF_AutoUPan) != 0)
				pannerCmp.uPanSpeed = zoneUPanSpeed;
			if((group.polyFlags & PF_AutoVPan) != 0)
				pannerCmp.vPanSpeed = zoneVPanSpeed;
			if((group.polyFlags & PF_SmallWavy) != 0)
				pannerCmp.wavy = true;
		}
		
		if ((group.polyFlags & PF_NotSolid) == 0)
			groupObj.AddComponent<MeshCollider> ();
		
		mesh.Optimize(); // do we need this?
	}

	public void Import(UnLevel target)
	{
		if (finishedImport)
			throw new Exception ("You can only import once per LevelImporting instance.");

		this.root = target;
		root.gameObject.isStatic = true;
		
		oldObjects = FindObjectsOfType<UnActor> ();

		// Remove old stuff
		for (int i=0; i<oldObjects.Length; i++) {
			if(oldObjects [i] != null)
			{
				if (oldObjects [i].originalLevel == root.gameObject && !(oldObjects[i] is UnZone) && !oldObjects[i].detached)
				{
					var oldGameObj = oldObjects[i].gameObject;
					DestroyImmediate (oldObjects [i]);
					DestroyImmediate (oldGameObj);
					oldObjects [i] = null;
				}
			}
		}
		for (int i=0; i<oldObjects.Length; i++) { // remove empty zones
			if(oldObjects [i] != null)
			{
				if (oldObjects [i].originalLevel == root.gameObject && oldObjects[i] is UnZone && !oldObjects[i].detached)
				{
					if(oldObjects[i].transform.childCount==0)
					{
						var oldGameObj = oldObjects[i].gameObject;
						DestroyImmediate (oldObjects [i]);
						DestroyImmediate (oldGameObj);
						oldObjects [i] = null;
					}
				}
			}
		}
		
		pkg = new PackageClasses.Package (srcFile);
		
		// Get model
		PackageClasses.ExportTableItem modelExp = Models.FindLevelModel (pkg);
		var model = Models.GetModel (modelExp);

		{
			PackageClasses.ExportTableItem levelInfo = null;
			foreach (var exp in pkg.ExportTable) {
				if (Utils.StrEq(exp.ClassNameStr, "LevelInfo")) {
					levelInfo = exp;
					break;
				}
			}
			zoneInfos[63] = levelInfo;
			
			for (int i=0; i<model.ZoneActors.Count; i++) {
				var act = model.ZoneActors[i];
				if(act==0)
					zoneInfos[i] = levelInfo;
				else
					zoneInfos[i] = pkg.ExportTable[act-1];
			}
		}
		
		for (int i=0; i<model.Nodes.Count; i++) {
			var n = model.Nodes[i];
			
			if(n.NumVertices>=3)
			{
				var surf = model.Surfs[n.iSurf];
				
				if((surf.PolyFlags & PF_FakeBackdrop) != 0 && noFakeBackdropPolys)
					continue;
				
				const int invisibleNonSolid = PF_Invisible | PF_NotSolid;
				if((surf.PolyFlags & invisibleNonSolid) != invisibleNonSolid)
				{
					var curPoly = new Polygon();
					
					curPoly.Flags = surf.PolyFlags;
					curPoly.Origin = VectorToUnity(model.Points[surf.pBase]);
					curPoly.Normal = VectorToUnity(model.Vectors[surf.vNormal]);
					curPoly.TextureU = VectorToUnity(model.Vectors[surf.vTextureU]);
					curPoly.TextureV = VectorToUnity(model.Vectors[surf.vTextureV]);
					curPoly.UPan = surf.PanU;
					curPoly.VPan = surf.PanV;
					
					curPoly.Texture = surf.Texture;
					curPoly.iZone = n.iZone1;
					
					curPoly.Verts = new Vector3[n.NumVertices];
					for(int j=n.iVertPool; j<n.iVertPool + n.NumVertices; j++)
						curPoly.Verts[j-n.iVertPool] = VectorToUnity(model.Points[model.Verts[j].pVertex]);
					
					AddBrushPolygon (curPoly);
				}
			}
		}
		
		zones.Sort ((x, y) => (x.zoneInfo.ObjectNameStr + " " + x.iZone).CompareTo (y.zoneInfo.ObjectNameStr + " " + y.iZone));
		
		for (int i=0; i<zones.Count; i++)
			FinishZone (zones [i]);

		// Import Lights, AmbientSounds and Movers...
		foreach (var exp in pkg.ExportTable) {
			ProcessExportItem(exp);
		}
		
		root.name = GetLevelTitle(pkg);

		var player = GameObject.FindWithTag ("Player");
		if (player != null)
		{
			var foundPlayerStart = false;

			foreach (var exp in pkg.ExportTable) {
				if (Utils.StrEq(exp.ClassNameStr, "PlayerStart")) {
					var props = exp.GetProperties ();
					
					if (props.GetBool ("bEnabled", true) && props.GetBool ("bSinglePlayerStart", true)) {
						player.transform.position = root.transform.position + SwapAxes(VectorToUnity(props.GetVector ("Location"))) * scale;
						player.transform.localEulerAngles = RotatorToUnity(props.GetRotator ("Rotation"));
						foundPlayerStart = true;
						break;
					}
				}
			}

			if(!foundPlayerStart)
				Debug.LogWarning("Level doesn't have proper PlayerStart.");
		}
		
		var skyCam = GameObject.Find ("Skybox Camera");
		if (skyCam != null)
		{
			PackageClasses.ExportTableItem skyZone = null;
			PackageClasses.ExportTableItem highDetailSkyZone = null;

			foreach (var exp in pkg.ExportTable) {
				if (Utils.StrEq(exp.ClassNameStr, "SkyZoneInfo")) {
					skyZone = exp;

					if(exp.GetProperties ().GetBool("bHighDetail"))
					{
						highDetailSkyZone = exp;
						break;
					}
				}
			}

			if(highDetailSkyZone != null)
				skyZone = highDetailSkyZone;

			if(skyZone != null)
			{
				var props = skyZone.GetProperties (); 
				skyCam.transform.position = root.transform.position + SwapAxes(VectorToUnity(props.GetVector ("Location"))) * scale;
			}
		}

		finishedImport = true;
	}

	void ProcessExportItem(PackageClasses.ExportTableItem exp)
	{
		if ((exp.ObjectFlags & (PackageClasses.EObjectFlags.RF_LoadForClient | PackageClasses.EObjectFlags.RF_LoadForServer)) == 0) //optimization
			return;

		var isLight = Utils.StrEq (exp.ClassNameStr, "Light");
		var isAmbientSound = Utils.StrEq(exp.ClassNameStr, "AmbientSound");
		var isMover = exp.ClassNameStr.ToLowerInvariant ().Contains ("mover");

		if (!isLight && !isAmbientSound && !isMover) //optimization
			return;

		//Don't reimport detached actors.
		// TODO: Iteration within iteration. May be slow.
		foreach(var oldObj in oldObjects)
			if (oldObj != null && oldObj.detached && oldObj.originalLevel == root.gameObject && Utils.StrEq(oldObj.originalName, exp.ObjectNameStr))
				return;

		var props = exp.GetProperties();

		var region = props.GetPointRegion ("Region");
		var parent = zones.Find (x => x.iZone == region.ZoneNumber).gameObject.transform;

		var movModelNum = props.GetInt("Brush");
		if (isMover && movModelNum == 0)
			return;

		var obj = new GameObject(exp.ObjectNameStr);
		obj.transform.SetParent(parent, false);

		if(isLight)
		{
			var lightType = (ELightType)(props.GetInt("LightType",1));
			var lightEffect = (ELightEffect)(props.GetInt("LightEffect",0));
			
			bool dynamicLight = true;
			
			switch(lightEffect)
			{
			case ELightEffect.None:
			case ELightEffect.CloudCast:
			case ELightEffect.Warp:
			case ELightEffect.StaticSpot:
			case ELightEffect.Spotlight:
			case ELightEffect.NonIncidence:
			case ELightEffect.Shell:
			case ELightEffect.OmniBumpMap:
			case ELightEffect.Sunlight:
				dynamicLight = false;
				break;
			}
			switch(lightType)
			{
			case ELightType.Pulse:
			case ELightType.Blink:
			case ELightType.Flicker:
			case ELightType.Strobe:
			case ELightType.SubtlePulse:
				dynamicLight = true;
				break;
			}

			var lightCmp = obj.AddComponent<UnLight>();

			var light = obj.GetComponent<Light>();
			light.shadows = LightShadows.Hard;
			light.range = props.GetInt("LightRadius",64) * 25f * scale;
			light.intensity = props.GetInt("LightBrightness",64) / 64f;//128f;
			light.color = HSVToRGB(255, props.GetInt("LightHue",0), props.GetInt("LightSaturation",255));
			//TODO: more LightTypes & LightEffects?
			if(!dynamicLight)
				light.renderMode = LightRenderMode.ForceVertex; //fixes double lighting on static lights

			switch(lightEffect)
			{
			case ELightEffect.Spotlight:
			case ELightEffect.StaticSpot:
			case ELightEffect.Searchlight:
				light.type = LightType.Spot;
				break;
			case ELightEffect.Sunlight:
				light.type = LightType.Directional;
				break;
			}
			
			light.spotAngle = props.GetInt ("LightCone", 128)*179f/255f;
			
			if(dynamicLight)
			{
				lightCmp.lightType   = lightType;
				lightCmp.lightEffect = lightEffect;
				lightCmp.lightPeriod = (byte)(props.GetInt ("LightPeriod",32));
				lightCmp.lightPhase  = (byte)(props.GetInt ("LightPhase",0));
				
				// hack to prevent baking dynamic lights
				lightCmp.intensity = light.intensity;
				light.intensity = 0;
			}

			obj.transform.localEulerAngles = RotatorToUnity(props.GetRotator("Rotation"));
		}
		else if(isAmbientSound)
		{
			obj.AddComponent<UnActor>();

			var soundCmp = obj.AddComponent<AudioSource>();
			soundCmp.loop = true;
			soundCmp.playOnAwake = true;
			soundCmp.spatialBlend = 1f;
			soundCmp.rolloffMode = AudioRolloffMode.Linear;
			var soundName = props["AmbientSound"];
			if(soundName != "")
			{
				soundCmp.clip = LoadSound(soundName);
				if(soundCmp.clip==null)
					Debug.LogWarning("Missing sound " + soundName);
			}
			soundCmp.pitch       = props.GetInt("SoundPitch" ,64 )/64f ;
			soundCmp.maxDistance = props.GetInt("SoundRadius",64 ) * 25f * scale;
			soundCmp.volume      = props.GetInt("SoundVolume",190)/255f;

			obj.transform.localEulerAngles = RotatorToUnity(props.GetRotator("Rotation"));
		}
		else if(isMover)
		{
			var tag = props["Tag"];
			var isTagSet = (tag != exp.ClassNameStr);

			obj.name += isTagSet ? " " + tag : "";

			var movModel = pkg.ExportTable[movModelNum - 1];
			var polysNum = Models.GetPolysNumFromModel(movModel);
			var polys = Models.GetPolys(pkg.ExportTable[polysNum - 1]);
			
			var mover = new Mover();
			
			movers.Add (mover);

			obj.isStatic = false;
			mover.gameObject = obj;

			obj.AddComponent<UnActor>();

			Vector3 prePivot = VectorToUnity(props.GetVector ("PrePivot"));
			Vector3 mainScale = VectorToUnity(props.GetScale ("MainScale").Scale);

			foreach(var poly in polys.PolyList)
			{
				var curPoly = new Polygon();
				
				curPoly.belongsToMover = true;
				
				curPoly.Flags = poly.Flags;
				curPoly.Origin = VectorToUnity (poly.Origin) - prePivot;
				curPoly.Normal = VectorToUnity (poly.Normal);
				curPoly.TextureU = VectorToUnity (poly.TextureU);
				curPoly.TextureV = VectorToUnity (poly.TextureV);
				curPoly.UPan = poly.UPan;
				curPoly.VPan = poly.VPan;
				
				curPoly.Texture = poly.Texture;
				
				curPoly.Verts = new Vector3[poly.VertList.Count];
				for(int i=0;i<poly.VertList.Count;i++)
					curPoly.Verts[i] = Vector3.Scale(VectorToUnity(poly.VertList[i]) - prePivot, mainScale);
				
				AddBrushPolygon (curPoly);
			}
			
			mover.polyGroups.Sort ((x, y) => x.textureName.CompareTo (y.textureName));
			
			foreach(var group in mover.polyGroups)
			{
				group.belongsToMover = true;
				FinishGroup (mover.gameObject, group);
			}

			obj.transform.localEulerAngles = ConvertRotator(RotatorToUnity(props.GetRotator ("Rotation")));
		}

		obj.transform.localPosition = SwapAxes(VectorToUnity(props.GetVector("Location"))) * scale;

		var actor = obj.GetComponent<UnActor>();
		actor.detached = false;
		actor.originalLevel = root.gameObject;
		actor.originalName = exp.ObjectNameStr;
		foreach(var prop in props.ToStringList())
			actor.properties.Add(prop.Name, prop.Value);
	}
}

#endif