﻿using UnityEngine;
using System.Collections;

[RequireComponent(typeof(Light))]
public class UnLight : UnActor {
	
	public ELightType lightType;
	[HideInInspector]
	public ELightEffect lightEffect;
	public byte lightPeriod;
	public byte lightPhase;

	new Light light;
	public float intensity = -1f;

	void Start () {
		light = GetComponent<Light> ();
		if (intensity < 0)
			intensity = light.intensity;
		else if(light.intensity==0) // hack to prevent baking dynamic lights
			light.intensity = intensity;
	}

	void Update () {
		switch (lightType) {
		case ELightType.Pulse:
			light.intensity = intensity * (Mathf.Sin ((Time.time+lightPhase/256f)*2f*Mathf.PI/(lightPeriod/32f))/2f+0.5f);
			break;
		case ELightType.SubtlePulse:
			light.intensity = intensity * (Mathf.Sin ((Time.time+lightPhase/256f)*2f*Mathf.PI/(lightPeriod/32f))/4f+0.75f);
			break;
		case ELightType.Strobe:
			if(light.intensity>0)
				light.intensity = 0;
			else
				light.intensity = intensity;
			break;
		case ELightType.Flicker:
			light.intensity = Random.value < 0.5 ? 0 : intensity;
			break;
		}

		switch (lightEffect) {
		case ELightEffect.Searchlight:
			transform.eulerAngles = new Vector3(0,(Time.time+lightPhase/256f)*360f/(lightPeriod/1.5f),0);
			break;
		}
	}
}

public enum ELightEffect
{
	None,
	TorchWaver,
	FireWaver,
	WateryShimmer,
	Searchlight,
	SlowWave,
	FastWave,
	CloudCast,
	StaticSpot,
	Shock,
	Disco,
	Warp,
	Spotlight,
	NonIncidence,
	Shell,
	OmniBumpMap,
	Interference,
	Cylinder,
	Rotor,
	Sunlight
}

public enum ELightType
{
	None,
	Steady,
	Pulse,
	Blink,
	Flicker,
	Strobe,
	BackdropLight,
	SubtlePulse,
	TexturePaletteOnce,
	TexturePaletteLoop
}